import collections

DEFAULT_ACTIONS = ['required', 'default', 'convert', 'type', 'valid', 'rename', ]
NONE_VALUES = [None, '', [], (), {}]
TYPE_MAP = {
    'int': int,
    'float': float,
    'number': (int, float),
    'str': str,
    'bool': bool,
    'list': (tuple, list, set),
    'dict': dict,
}


class PurifyException(Exception):
    def __init__(self, message):
        self.message = message


class PurifyActionException(Exception):
    def __init__(self, key, value, action=None, param=None, message=None):
        self.action = action
        self.key = key
        self.value = value
        self.param = param
        self.message = message


class skip:
    """
    When null is this class, that will remove in value dict.
    """
    pass


class Hyena:
    """
    example of inspector map

    ALLOWED_ACTIONS: required, type, default, convert, valid, rename

    {
        key: {
            'required': True or False,
            'type': ['int', 'float', 'number', 'str', 'list', 'dict', 'bool'],
            'default': {DEFAULT_VALUE},
            'convert': [int, float, str, list, convert_function, ...],
            'valid': {Validation Function},
            'rename': string
        },
        another_key: {
            ...actions
        }
    }

    # valid function example
    def valid_func(value, key, schema):
        return True or False

    # rename function example
    def rename_func(value, original):
        # original: original key
        return 'new_name'

    """
    type_maps = TYPE_MAP
    none_values = NONE_VALUES
    allow_actions = DEFAULT_ACTIONS

    def __init__(self, schema=None):
        self.schema = schema

    def action_unknown(self, key, value, param=None):
        """
        When data not exists in inspect_map
        """
        return skip

    def action_required(self, key, value, param):
        """
        When field has required.
        """
        required = param
        if required and value in self.none_values:
            raise PurifyActionException(
                key, value, action='required', param=param,
                message='required field {} is not exists'.format(key)
            )
        return value

    def action_convert(self, key, value, param):
        """
        When field has convert, convert value
        """
        if not callable(param):
            raise PurifyException(
                message='{} convert param is not callable'.format(key)
            )
        return param(value)

    def action_type(self, key, value, param):
        """
        When field has type, check type.
        """
        if param not in self.type_maps:
            raise PurifyActionException(
                key, value, action='type', param=param,
                message='type: {} is not allowed'.format(param)
            )
        if value is None:
            return value

        standard = self.type_maps[param]
        value_type = type(value)

        if value_type is standard:
            return value
        if isinstance(standard, collections.Iterable) and value_type in standard:
            return value
        raise PurifyActionException(
            key, value, action='type', param=param,
            message='value type: {} is not {}'.format(value_type, param)
        )

    def action_default(self, key, value, param):
        """
        When field has default and value in none_values, return default
        """
        default = param
        if value in self.none_values:
            if param is skip:
                return skip
            if callable(default):
                return default()
            return default
        return value

    def action_valid(self, key, value, param):
        if not callable(param):
            raise PurifyActionException(
                key, value, action='valid', param=param,
                message='valid function: {} is not callable'.format(param)
            )

        if param(value, key=key, schema=self.schema):
            return value
        raise PurifyActionException(
            key, value, action='valid', param=param,
            message='validation error: {} is invalid value {}'.format(key, value)
        )

    def action_rename(self, key, value, param):
        if callable(param):
            return param(value, original=key)
        return param

    def get_clear_field(self, field):
        if 'required' in field and 'default' in field:
            raise PurifyException(
                'Purify failed: required action cannot be used with default.'
            )
        if 'convert' in field and 'type' in field:
            raise PurifyException(
                'Purify failed: convert action cannot be used with type.'
            )

        result = {}
        for action in self.allow_actions:
            if action in field:
                result[action] = field[action]
        return result

    def action(self, action, key, value, param):
        if action == 'unknown':
            return self.action_unknown(key, value, param)
        if action not in self.allow_actions:
            raise PurifyException('invalid action: {}'.format(action))
        method = getattr(self, 'action_{}'.format(action))
        return method(key, value, param)

    def purify_field(self, key, value, field):
        cleared_field = self.get_clear_field(field)
        # rename action always run first
        if 'rename' in field:
            key = self.action_rename(key, value, cleared_field.pop('rename'))

        for action, param in cleared_field.items():
            value = self.action(action, key, value, param)
            if value is skip:
                break
        return key, value

    def purify(self, data, schema=None):
        if schema:
            self.schema = schema
        if self.schema is None:
            raise PurifyException('Purify failed: no schema')

        result = {}
        # field in schema
        for key, field in self.schema.items():
            value = data.get(key)
            key, value = self.purify_field(key, value, field)
            if value is skip:
                continue
            result[key] = value

        # data not in schema
        unknown_data = {key: value for key, value in data.items() if key not in self.schema}
        for key, value in unknown_data.items():
            value = self.action('unknown', key, value, None)
            if value is skip:
                continue
            result[key] = value

        return result



from unittest import TestCase

from hyena import Hyena, PurifyActionException, skip


def test_validation1(value, **kwargs):
    if type(value) is not list:
        return False
    return len(value) >= 1


def test_validation2(value, **kwargs):
    if type(value) is not dict:
        return False
    return 'field' in value


class TestInspector(TestCase):
    def test_unknown(self):
        target = {
            'one': 1,
            'two': 2,
            'three': 3,
            'four': 4,
        }
        schema = {
            'one': {},
            'three': {},
            'four': {}
        }
        expected = {
            'one': 1,
            'three': 3,
            'four': 4,
        }
        hyn = Hyena(schema)
        result = hyn.purify(target)
        self.assertEqual(expected, result)

    def test_required(self):
        def _required_checker(key='three'):
            with self.assertRaises(PurifyActionException) as res:
                hyn.purify(target)
            self.assertEqual(key, res.exception.key)
            self.assertEqual('required', res.exception.action)

        target = {
            'one': 0,
            'two': False,
            'three': 3,
            'four': 4,
        }
        schema = {
            'one': {'required': True},
            'two': {'required': True},
            'three': {'required': True},
            'four': {},
        }
        hyn = Hyena(schema)

        for i in ['', None, [], {}, ()]:
            target.update({'three': i})
            _required_checker()

        target.pop('three')
        _required_checker()

        target.update({'three': 3})
        result = hyn.purify(target)
        self.assertEqual(target, result)

    def test_convert(self):
        def _inner_convertor(value):
            return value.split(',')

        target = {
            'one': 123,
            'two': '1234',
            'three': 'one,two,three',
            'four': 0,
            'five': None,
        }
        schema = {
            'one': {'convert': str},
            'two': {'convert': int},
            'three': {'convert': _inner_convertor},
            'four': {'convert': bool},
            'five': {'default': skip, 'convert': int},  # if skip not work, error raised.
        }
        expected = {
            'one': '123',
            'two': 1234,
            'three': ['one', 'two', 'three'],
            'four': False,
        }

        hyn = Hyena(schema)
        result = hyn.purify(target)
        self.assertEqual(expected, result)

    def test_type(self):
        def _type_checker(_key):
            with self.assertRaises(PurifyActionException) as res:
                hyn.purify(target)
            self.assertEqual(_key, res.exception.key)
            self.assertEqual('type', res.exception.action)
            self.assertEqual(_key, res.exception.param)

        target = {
            'int': 123,
            'str': '123',
            'float': 0.123,
            'number': 123,
            'bool': False,
            'list': [1, 2, 3],
            'dict': {'key': 'value'}
        }
        schema = {
            'int': {'type': 'int'},
            'str': {'type': 'str'},
            'float': {'type': 'float'},
            'number': {'type': 'number'},
            'bool': {'type': 'bool'},
            'list': {'type': 'list'},
            'dict': {'type': 'dict'},
        }
        hyn = Hyena(schema)
        result = hyn.purify(target)
        self.assertEqual(target, result)

        check_list = [
            ('int', 'string'),
            ('str', 123),
            ('float', 'string'),
            ('number', 'string'),
            ('bool', [1, 2, 3]),
            ('list', {'key': 'value'}),
            ('dict', 123),
        ]

        for key, value in check_list:
            target.update({key: value})
            _type_checker(key)
            target.pop(key)

    def test_default(self):
        target = {
            'one': '',
            'two': None,
            'three': {},
            'four': [],
            'five': 'value',
        }
        schema = {
            'one': {'default': skip},
            'two': {'default': None},
            'three': {'default': 'default'},
            'four': {'default': 100},
            'five': {},
            'six': {'default': 'NOT_EXISTS_DATA'},
        }
        expected = {
            'two': None,
            'three': 'default',
            'four': 100,
            'five': 'value',
            'six': 'NOT_EXISTS_DATA'
        }

        hyn = Hyena(schema)
        result = hyn.purify(target)
        self.assertEqual(expected, result)

    def test_valid(self):
        def _check_raise(_target, key):
            with self.assertRaises(PurifyActionException) as e:
                hyn.purify(_target)
                exc = e.exception
                self.assertEqual(exc.action, 'valid')
                self.assertEqual(exc.key, key)

        schema = {
            'one': {'valid': test_validation1},
            'two': {'valid': test_validation2},
        }
        target = {'one': 'NOT_VALID'}

        hyn = Hyena(schema)
        _check_raise(target, 'one')

        target['one'] = []
        _check_raise(target, 'one')

        target['one'].append(1)
        target['two'] = {}
        _check_raise(target, 'two')

        # Success
        target['two']['field'] = 1
        hyn.purify(target)

    def test_rename(self):
        def rename_func(value, original):
            return value + '_' + original

        schema = {
            'oneChanged': {'rename': 'one_changed'},
            'with_func': {'rename': rename_func}
        }
        target = {
            'oneChanged': 'value',
            'with_func': 'rename',
        }
        expect = {
            'one_changed': 'value',
            'rename_with_func': 'rename',
        }

        hyn = Hyena(schema)
        result = hyn.purify(target)
        self.assertEqual(expect, result)

    def test_purify(self):
        target = {
            'one': 123,
            'two': None,
            'three': 123,
        }
        schema = {
            'one': {'required': True, 'type': 'int'},
            'two': {'type': 'str', 'default': 'default_str'},
            'three': {'required': True, 'convert': str},
        }
        expected = {
            'one': 123,
            'two': 'default_str',
            'three': '123',
        }

        hyn = Hyena(schema)
        result = hyn.purify(target)
        self.assertEqual(expected, result)



# -*- coding: utf-8 -*-
from setuptools import setup

modules = \
['hyena']
setup_kwargs = {
    'name': 'hyena-py',
    'version': '0.0.1',
    'description': 'data purifier with schema',
    'long_description': '[![version badge](https://badge.fury.io/py/hyena.svg)](https://badge.fury.io/py/hyena)\n\n**hyena** is data purifier with schema\n\n## Warning\n\nThis repository is on developing. Some bugs may exist, possible to change suddenly. And no document yet.\n\n\n## Installation\n\n    pip install hyena-py\n',
    'author': 'jrog612',
    'author_email': 'jrog612@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/jrog612/hyena',
    'py_modules': modules,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
